from dataclasses import dataclass
from datetime import date

import pandas as pd


@dataclass
class ParsedData:
    """
    Dataclass to store parsed data from the Apple Health export file.
    """

    flag: str
    sources: list[str]
    dates: set[date]
    records: pd.DataFrame

    def __str__(self) -> str:
        """
        String representation of the ParsedData class.
        Includes the flag, dates, and number of records.

        Example:
        ```bash
        =====================ParsedData=====================
        Flag:       HKQuantityTypeIdentifierRestingHeartRate
        Sources:    3 sources
        Dates:      144 dates
        Records:    145 records
        ```

        Returns:
            str: String representation of the ParsedData class
        """

        description = [
            f"{'Flag:':<12}{self.flag}",
            f"{'Sources:':<12}{len(self.sources)} sources",
            f"{'Dates:':<12}{len(self.dates)} dates",
            f"{'Records:':<12}{len(self.records)} records",
        ]

        max_len = len(max(description, key=len))
        description = [f"{'ParsedData':=^{max_len}}"] + description

        return "\n".join(description)



from apple_health_parser.consts import ALLOWED_IMAGE_FORMATS, OPERATIONS


class MissingFlag(Exception):
    """
    Exception to raise when the flag is missing.
    """

    def __init__(self, flag: str) -> None:
        super().__init__(f"Flag '{flag}' is missing in the data.")


class MissingRecords(Exception):
    """
    Exception to raise when records are missing.
    """

    def __init__(self) -> None:
        super().__init__("No records found in the data.")


class MissingYear(Exception):
    """
    Exception to raise when there is no record for the given year.
    """

    def __init__(self, year: int, years: list[int]) -> None:
        super().__init__(f"No record for the year {year}. Available years: {years}.")


class InvalidFileFormat(Exception):
    """
    Exception to raise when the file type is incorrect.
    """

    def __init__(self, format: str) -> None:
        super().__init__(
            f"File type is incorrect. Requires .csv, but provided {format}."
        )


class InvalidFlag(Exception):
    """
    Exception to raise when the flag is incorrect.
    """

    def __init__(self, flag: str, flags: list[str]) -> None:
        super().__init__(f"Flag '{flag}' is not valid. Allowed flags: {flags}.")


class InvalidHeatmapOperation(Exception):
    """
    Exception to raise when the operation is invalid for a heatmap (i.e. when operation is `None`).
    """

    def __init__(self) -> None:
        super().__init__(
            f"Heatmaps require an operation. Allowed operations: {OPERATIONS}."
        )


class InvalidImageFormat(Exception):
    """
    Exception to raise when the image format is incorrect.
    """

    def __init__(self, format: str) -> None:
        super().__init__(
            f"Incorrect image format: '{format}'. Allowed formats: {ALLOWED_IMAGE_FORMATS}."
        )


class InvalidOperation(Exception):
    """
    Exception to raise when the operation is invalid.
    """

    def __init__(self, operation: str | None) -> None:
        super().__init__(
            f"Operation '{operation}' is invalid. Allowed operations: {OPERATIONS}."
        )


class InvalidSource(Exception):
    """
    Exception to raise when the source is missing.
    """

    def __init__(self, source: str, sources: list[str]) -> None:
        super().__init__(
            f"No records found for '{source}'. Available sources: {sources}."
        )



from abc import ABC, abstractmethod
from datetime import date
from pathlib import Path

from plotly.graph_objs import Figure

from apple_health_parser.config.definitions import AllowedImageFormats, PlotSettings
from apple_health_parser.consts import Operations, PlotType
from apple_health_parser.exceptions import InvalidImageFormat
from apple_health_parser.models.parsed import ParsedData
from apple_health_parser.utils.preprocessor import Preprocessor


class PlotInterface(ABC):
    def __init__(
        self,
        data: ParsedData,
        year: int = date.today().year,
        source: str | None = None,
        operation: str | None = None,
        heatmap: bool = False,
        title: bool = False,
    ) -> None:
        """
        Initialize the Plot object.

        Setup the plot after preprocessing the data.

        Args:
            data (ParsedData): ParsedData object
            year (int, optional): Year, defaults to date.today().year
            source (str | None, optional): Source, defaults to None
            operation (str | None, optional): Operation, defaults to None
            heatmap (bool, optional): Flag to plot a heatmap, defaults to False
            title (bool, optional): Flag to include the plot title, defaults to False
        """
        self.src = source
        self.data = data
        self.year = year
        self.oper = operation
        self.hmap = heatmap
        self.title = title
        self.flag = data.flag

        preprocessor = Preprocessor(
            data=data, year=year, source=source, operation=operation, heatmap=heatmap
        )
        self.dataframe = preprocessor.get_dataframe()
        self.meta = preprocessor.meta

        # Plot type after preprocessor due to validity checks in preprocessor
        self.ptype: PlotType = self._get_plot_type()
        self.psets: PlotSettings = self._get_plot_settings()

    def _get_plot_settings(self) -> PlotSettings:
        """
        Get the plot settings.

        For the heart rate flag, the x-axis is "start_date" and the y-axis is "value".
        It also includes the color as "motion_context" and the legend as "Motion Context".

        Returns:
            PlotSettings: Plot settings
        """
        color = None
        colormap = None
        legend = None
        title = None
        title_yaxis = None

        match self.flag:
            case "HKQuantityTypeIdentifierHeartRate":
                x = "start_date"
                y = "value"
                color = "motion_context"
                legend = "Motion Context"
            case _:
                x = "date"
                y = "value"

        match self.title:
            case True:
                title = (
                    f"{self.meta.name} {self.year}"
                    f"{' (' + self.oper + ')' if self.oper else ''}"
                    f"{': ' + self.src if self.src else ''}"
                )

        match self.ptype:
            case PlotType.HEATMAP:
                colormap = self.meta.colormap
            case _:
                title_yaxis = f"{self.meta.name} {self.meta.unit} ({self.oper})"

        return PlotSettings(x, y, color, colormap, legend, title, title_yaxis)

    def _get_plot_type(self) -> PlotType:
        """
        Get the plot type based on the operation and whether a heatmap is requested.

        Returns:
            PlotType: Plot type
        """
        match self.hmap:
            case True:
                return PlotType.HEATMAP
            case False:
                match self.oper:
                    case Operations.COUNT | Operations.SUM:
                        return PlotType.BAR
                    case (
                        Operations.MAX
                        | Operations.MEAN
                        | Operations.MEDIAN
                        | Operations.MIN
                    ):
                        return PlotType.LINE
                    case _:
                        return PlotType.SCATTER

    @abstractmethod
    def _get_figure(self) -> Figure:
        pass

    def plot(
        self, show: bool = True, save: bool = False, format: str = "png"
    ) -> Figure:
        """
        Plot the figure.

        Args:
            show (bool, optional): Whether to show the figure, defaults to True
            save (bool, optional): Whether to save the figure, defaults to False
            format (str, optional): Format of the file to be saved, defaults to "png"

        Raises:
            InvalidImageFormat: Image format is not allowed

        Returns:
            Figure: Figure object
        """
        figure = self._get_figure()

        if show:
            figure.show()

        if save:
            if format.upper() not in AllowedImageFormats.__members__:
                raise InvalidImageFormat(format)

            output_dir = Path("plots")
            output_dir.mkdir(exist_ok=True)
            lowercase_flag = self.meta.name.replace(" ", "_").lower()
            filename = f"{'heatmap' if self.hmap else 'plot'}_{lowercase_flag}_{self.year}{'_' + self.oper if self.oper else ''}"

            if format == "html":
                figure.write_html(output_dir / f"{filename}.html")
            else:
                figure.write_image(
                    file=output_dir / f"{filename}.{format}", format=format, scale=1
                )

        return figure



from datetime import datetime
from enum import StrEnum

from pydantic import BaseModel, Field, field_validator


class MotionContext(StrEnum):
    UNSET = "0"
    SEDENTARY = "1"
    ACTIVE = "2"


class HealthData(BaseModel):
    type: str = Field(
        alias="type",
        title="Type",
        description="Type of health data",
        examples=["HKQuantityTypeIdentifierHeartRate"],
    )
    source_name: str = Field(
        alias="sourceName",
        title="Source Name",
        description="Name of the source",
        examples=["Apple Watch"],
    )
    source_version: str = Field(
        alias="sourceVersion",
        title="Source Version",
        description="Version of the source",
        examples=["10.2"],
    )
    unit: str = Field(
        alias="unit",
        title="Unit",
        description="Unit of the health data",
        examples=["count/min", "kcal", "%"],
    )
    creation_date: datetime = Field(
        alias="creationDate",
        title="Creation Date",
        description="Date of creation",
        examples=["2021-01-01 00:00:00 +0200"],
    )
    start_date: datetime = Field(
        alias="startDate",
        title="Start Date",
        description="Date of measurement start",
        examples=["2021-01-01 00:00:00 +0200"],
    )
    end_date: datetime = Field(
        alias="endDate",
        title="End Date",
        description="Date of measurement end",
        examples=["2021-01-01 00:00:00 +0200"],
    )
    value: int | float = Field(
        alias="value",
        title="Value",
        description="Value of the health data",
        examples=[60, 120, 15.5],
    )

    @field_validator("creation_date", "start_date", "end_date", mode="before")
    @classmethod
    def check_date(cls, v: str) -> datetime:
        return datetime.strptime(v, "%Y-%m-%d %H:%M:%S %z")

    @field_validator("value", mode="before")
    @classmethod
    def str_to_numeric(cls, v) -> int | float:
        if type(v) is str:
            try:
                return int(v)
            except ValueError:
                return float(v)
        return v


class HeartRateData(HealthData):
    device: str = Field(title="Device", description="Device used for measurement")
    motion_context: str = Field(
        alias="motionContext",
        title="Motion Context",
        description="Context of motion (e.g. sedentary, active, unset)",
        examples=["0", "1", "2"],
    )

    @field_validator("motion_context", mode="before")
    @classmethod
    def check_motion_context(cls, v: str) -> str:
        return MotionContext(v).name.lower().capitalize()



import re
from pathlib import Path
from shutil import rmtree
from xml.etree import ElementTree as ET
from zipfile import ZipFile

import click

from apple_health_parser.utils.logging import logger


class Loader:
    """
    Loader class to extract and read an XML file from the Apple Health `export.zip` file.
    """

    @staticmethod
    def extract_zip(
        zip_file: str | Path, output_dir: str | Path, overwrite: bool | None = None
    ) -> Path:
        """
        Extracts a zip file to an output directory.

        Args:
            zip_file (str | Path): The zip file to extract
            output_dir (str | Path): The output directory to extract the file to
            overwrite (bool, optional): Flag to overwrite the existing data, defaults to None

        Returns:
            Path: The absolute path to the extracted file
        """
        if isinstance(zip_file, str):
            zip_file = Path(zip_file)

        if isinstance(output_dir, str):
            output_dir = Path(output_dir)

        export_dir = output_dir / "apple_health_export"

        # Check if output directory exists and delete it if it does and "y" or "yes" is entered
        if export_dir.exists():
            Loader.delete_previous_export(export_dir, overwrite)

        # Extract the zip file
        with ZipFile(zip_file, "r") as data:
            logger.info(f"Extracting {zip_file} to {output_dir}...")
            data.extractall(output_dir)

        # Log the compressed and uncompressed file sizes
        file_size = zip_file.stat().st_size
        dir_size = sum(f.stat().st_size for f in output_dir.glob("**/*") if f.is_file())
        logger.info(f"Compressed: {file_size / 1e6:.2f} MB")
        logger.info(f"Uncompressed: {dir_size/ 1e6:.2f} MB")

        return (export_dir / "export.xml").resolve()

    @staticmethod
    def delete_previous_export(output_dir: Path, overwrite: bool | None) -> None:
        """
        Delete the previous export if it exists and the user agrees.

        Args:
            output_dir (Path): The output directory to extract the file to
            overwrite (bool | None): Flag to overwrite the existing data, defaults to None
        """
        match overwrite:
            case None:
                logger.warning(f"Found previous export at {output_dir}...")
                if click.confirm("Do you want to delete it?"):
                    rmtree(output_dir)
                    logger.warning(f"Deleted previous export at {output_dir}...")

            case True:
                rmtree(output_dir)
                logger.warning(f"Deleted previous export at {output_dir}...")

    @staticmethod
    def read_xml(xml_file: Path) -> list[ET.Element]:
        """
        Read an XML file and return the root element.

        Args:
            xml_file (Path): Path to the XML file

        Returns:
            list[ET.Element]: List of records (ET.Element)
        """
        logger.info(f"Processing {xml_file}...")
        with open(xml_file, "r") as file:
            root = ET.parse(file).getroot()
            Loader._log_metadata(root)
            return root.findall("Record")

    @staticmethod
    def _log_metadata(root: ET.Element) -> None:
        """
        Log metadata from the XML file.

        Example:

        ```bash
        2024-05-29 22:28:23,064 - INFO - Locale:                            en_NL
        2024-05-29 22:28:23,065 - INFO - Export date:                       2024-05-30 22:20:25 +0200
        2024-05-29 22:28:23,065 - INFO - Date of birth:                     1990-04-22
        2024-05-29 22:28:23,065 - INFO - Biological sex:                    HKBiologicalSexMale
        2024-05-29 22:28:23,065 - INFO - Blood type:                        HKBloodTypeAPositive
        2024-05-29 22:28:23,065 - INFO - Fitzpatrick skin type:             HKFitzpatrickSkinTypeNotSet
        2024-05-29 22:28:23,065 - INFO - Cardio fitness medications use:    None
        ```

        Args:
            root (ET.Element): Root element of the XML file
        """

        def get_locale() -> None:
            """
            Get the locale from the XML file (e.g. `en_NL`).
            """
            locale = root.attrib.get("locale")
            if locale is not None:
                logger.info(f"{'Locale:':<35}" + click.style(f"{locale}", bg="red"))

        def get_export_date() -> None:
            """
            Get the export date from the XML file (e.g. `2024-05-29 22:20:35 +0200`).
            """
            export_date = root.find("ExportDate")
            if export_date is not None:
                logger.info(
                    f"{'Export date:':<35}"
                    + click.style(f"{export_date.attrib['value']}", fg="green")
                )

        def get_user() -> None:
            """
            Get user metadata from the XML file.
            """
            user = root.find("Me")
            if user is not None:
                user_data: list[tuple[str, str]] = [
                    (flag.removeprefix("HKCharacteristicTypeIdentifier"), value)
                    for (flag, value) in user.items()
                ]
                for flag, value in user_data:
                    # Split flag at uppercase letters
                    flag = " ".join(re.findall(r"[A-Z][^A-Z]*", flag))
                    # Lower case every word except the first
                    flag = flag[0] + flag[1:].lower() + ":"
                    logger.info(f"{flag:<35}" + click.style(f"{value}", fg="cyan"))

        get_locale()
        get_export_date()
        get_user()



import click

from apple_health_parser.plot.overviews import ActivityOverview
from apple_health_parser.plot.plots import Plot
from apple_health_parser.utils.logging import logger
from apple_health_parser.utils.parser import Parser


@click.command(no_args_is_help=True)
@click.option("--zip_file", help="Path to the Apple Health export.zip file")
def main(zip_file: str) -> None:
    """
    CLI to parse and plot data from the Apple Health export file.

    Args:
        zip_file (str): Path to the export.zip file
    """
    logger.info(click.style("Apple Health Parser", bg="blue", fg="white", bold=True))

    parser = Parser(export_file=zip_file, overwrite=True, verbose=True)

    # Get the sources from the export file
    flag = "HKQuantityTypeIdentifierDistanceWalkingRunning"
    source = "Alexandre’s Apple\xa0Watch"
    data = parser.get_flag_records(flag=flag)

    plot = Plot(data=data, source=source, operation="sum", heatmap=False)
    plot.plot(show=True, save=True, format="svg")

    # Activity overview
    flags = [
        "HKQuantityTypeIdentifierActiveEnergyBurned",
        "HKQuantityTypeIdentifierAppleExerciseTime",
        "HKQuantityTypeIdentifierAppleStandTime",
    ]
    data = parser.get_flag_records(flag=flags)
    overview = ActivityOverview(data=data, year=2024, source=source)
    overview.plot(show=True, save=True, format="svg")


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod

import pandas as pd

from apple_health_parser.config.definitions import Metadata
from apple_health_parser.consts import FLAG_METADATA, OPERATIONS
from apple_health_parser.exceptions import (
    InvalidHeatmapOperation,
    InvalidOperation,
    InvalidSource,
    MissingFlag,
    MissingYear,
)
from apple_health_parser.models.parsed import ParsedData


class PreprocessorInterface(ABC):
    def __init__(
        self,
        data: ParsedData,
        year: int,
        source: str | None = None,
        operation: str | None = None,
        heatmap: bool = False,
    ) -> None:
        """
        Initialize the Preprocessor object.

        Prepare the DataFrame for the plot.

        Args:
            data (ParsedData): ParsedData object
            year (int, optional): Year, defaults to date.today().year
            source (str | None, optional): Source, defaults to None
            operation (str | None, optional): Operation, defaults to None
            heatmap (bool, optional): Flag to plot a heatmap, defaults to False
        """
        self.src = source
        self.data = data
        self.year = year
        self.oper = operation
        self.hmap = heatmap
        self.flag = data.flag
        self._validate()

    @property
    def meta(self) -> Metadata:
        """
        Metadata for the flag.

        Raises:
            MissingFlag: Missing flag metadata

        Returns:
            Metadata: Metadata for the flag
        """
        metadata = FLAG_METADATA.get(self.flag)
        if metadata is None:
            raise MissingFlag(self.flag)
        return metadata

    def _validate(self) -> None:
        """
        Validate the flag, operation, and year.

        Raises:
            MissingYear: Missing year in the records
            InvalidOperation: Invalid operation
            InvalidHeatmapOperation: Invalid operation for heatmap
            InvalidSource: Invalid source name
        """
        years = self.data.records.start_date.dt.year.unique().tolist()
        if self.year not in years:
            raise MissingYear(self.year, years)

        if self.oper is not None:
            if self.oper not in OPERATIONS:
                raise InvalidOperation(self.oper)
        else:
            if self.hmap:
                raise InvalidHeatmapOperation

        sources = self.data.records.source_name.unique()
        if self.src and self.src not in sources:
            raise InvalidSource(self.src, sources)

    @abstractmethod
    def get_heatmap(self, data: pd.DataFrame) -> pd.DataFrame:
        pass

    @abstractmethod
    def get_dataframe(self) -> pd.DataFrame:
        pass



import time
from functools import wraps

from apple_health_parser.utils.logging import logger


def timeit(func):
    """
    Log (`log.info`) how long a function takes to run (in seconds).
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        logger.info(f"Took {time.time() - start:.2f} seconds to run {func.__name__}")
        return result

    return wrapper



from dataclasses import dataclass
from enum import StrEnum
from importlib import resources

import yaml


@dataclass
class Metadata:
    """
    Represents metadata for a specific data type in Apple Health.

    Attributes:
        name (str): The name of the data type
        unit (str): The unit of measurement for the data type
        color (str): The color associated with the data type
        colormap (str): The colormap associated with the data type
    """

    name: str
    unit: str
    color: str
    colormap: str


def get_flag_metadata() -> dict[str, Metadata]:
    """
    Get the metadata for each flag.

    Returns:
        dict[str, Metadata]: Dictionary of flag metadata with flag as key and Metadata as value
    """
    flags_file = resources.files("apple_health_parser.config") / "flags.yaml"
    flags_content = yaml.safe_load(flags_file.read_text())
    return {flag: Metadata(**metadata) for flag, metadata in flags_content.items()}


class AllowedImageFormats(StrEnum):
    HTML = "html"
    PNG = "png"
    JPEG = "jpeg"
    WEBP = "webp"
    SVG = "svg"
    PDF = "pdf"


class Operations(StrEnum):
    COUNT = "count"
    MAX = "max"
    MEAN = "mean"
    MEDIAN = "median"
    MIN = "min"
    SUM = "sum"


class PlotType(StrEnum):
    BAR = "bar"
    LINE = "line"
    SCATTER = "scatter"
    HEATMAP = "imshow"


@dataclass
class PlotSettings:
    """
    Represents the settings for a plot.

    Attributes:
        x (str): The column name to be plotted on the x-axis
        y (str): The column name to be plotted on the y-axis
        color (str | None): The variable to determine the color of the plot (optional)
        colormap (str | None): The colormap to be used for coloring the plot (optional)
        legend (str | None): The title of the legend (optional)
        title (str | None): The title of the plot (optional)
        title_yaxis (str | None): The title of the y-axis (optional)
    """

    x: str
    y: str
    color: str | None
    colormap: str | None
    legend: str | None
    title: str | None
    title_yaxis: str | None



import plotly.express as px
from plotly.graph_objs import Figure

from apple_health_parser.consts import PlotType
from apple_health_parser.interfaces.plot_interface import PlotInterface


class Plot(PlotInterface):
    """
    Plot the parsed data.
    """

    def _get_figure(self) -> Figure:
        """
        Get the plotly figure:
        - If the operation is not provided, the plot type is `PlotType.SCATTER`
        - If the operation is "count" or "sum", the plot type is `PlotType.BAR`
        - If the operation is "max", "mean", "median", or "min", the plot type is `PlotType.LINE`

        Returns:
            Figure: Figure object
        """
        if self.ptype is PlotType.HEATMAP:
            fig: Figure = px.imshow(
                self.dataframe,
                labels=dict(
                    x="Day of the month",
                    y="Month",
                ),
                text_auto=True,
                color_continuous_scale=self.psets.colormap,
                title=self.psets.title,
                template="simple_white",
            )

            fig.update_layout(
                coloraxis=dict(colorbar=dict(orientation="h", y=-0.15)),
                width=720,
                # This is a workaround to fix the height of the heatmap based on the number of months
                height=150
                + 30 * (self.dataframe.index.max() - self.dataframe.index.min() + 1),
            )

            fig.update_coloraxes(showscale=False)
            fig.update_xaxes(
                title="Day of the Month",
                tickvals=list(range(1, 32)),
                showline=True,
                linewidth=1,
                linecolor="black",
                mirror=True,
            )
            fig.update_yaxes(
                title="Month",
                tickvals=list(
                    range(
                        self.dataframe.index.min(),
                        self.dataframe.index.max() + 1,
                    )
                ),
                showline=True,
                linewidth=1,
                linecolor="black",
                mirror=True,
            )

        else:
            fig: Figure = getattr(px, self.ptype)(
                data_frame=self.dataframe,
                x=self.psets.x,
                y=self.psets.y,
                color=self.psets.color,
                title=self.psets.title,
                template="simple_white",
            )

            if self.ptype is PlotType.SCATTER:
                fig.update_traces(marker=dict(size=4))

            elif self.ptype is PlotType.BAR:
                fig.update_coloraxes(showscale=False)

            fig.update_layout(
                xaxis_title="Date",
                yaxis_title=self.psets.title_yaxis,
                legend_title_text=self.psets.legend,
            )

        return fig



import xml.etree.ElementTree as ET
from datetime import date
from pathlib import Path

import click
import pandas as pd

from apple_health_parser.decorators import timeit
from apple_health_parser.exceptions import (
    InvalidFileFormat,
    InvalidFlag,
    MissingRecords,
)
from apple_health_parser.models.parsed import ParsedData
from apple_health_parser.models.records import HealthData, HeartRateData
from apple_health_parser.utils.loader import Loader
from apple_health_parser.utils.logging import logger

# https://pandas.pydata.org/pandas-docs/stable/user_guide/indexing.html#returning-a-view-versus-a-copy
pd.options.mode.copy_on_write = True


class Parser(Loader):
    """
    Parser class to parse the Apple Health export file.

    Subclass of the Loader class which extracts the export.zip file and logs metadata.
    """

    def __init__(
        self,
        export_file: str | Path,
        output_dir: str | Path = "data",
        verbose: bool = False,
        overwrite: bool | None = None,
    ) -> None:
        """
        Initialize the Parser class with the path to the export.zip file.

        Args:
            export_file (str): Path to the export.zip file
            output_dir (str): Directory to export the parsed data to, defaults to "data"
            verbose (bool): Flag to enable verbose logging, defaults to False
            overwrite (bool, optional): Flag to overwrite the existing data, defaults to None
        """
        if verbose is False:
            logger.propagate = False

        self.xml_file = self.extract_zip(
            zip_file=export_file, output_dir=output_dir, overwrite=overwrite
        )
        self.records = self._get_records()
        self.sources = self.get_sources()

    @timeit
    def _get_records(self) -> dict[str, list[ET.Element]]:
        """
        Get records from the Apple Health export file.
        The records are grouped by flags as keys and list of records as values.

        Returns:
            dict[str, list[ET.Element]]: Records from the export.xml file
        """
        data = self.read_xml(self.xml_file)

        self.flags = self._get_flags(data)
        records = {
            flag: [rec for rec in data if rec.attrib["type"] == flag]
            for flag in self.flags
        }

        record_count = sum(len(rec) for rec in records.values())
        logger.info(
            f"Processed {len(records.keys())} flags with {record_count:,} records"
        )

        return records

    def _get_flags(self, data: list[ET.Element]) -> list[str]:
        """
        Get flags from the Apple Health records.

        Args:
            data (list[ET.Element]): List of records from the export.xml file

        Returns:
            list[str]: Sorted list of flags
        """
        return sorted({rec.attrib["type"] for rec in data})

    def _build_models(self, flag: str) -> list:
        """
        Build models from the records based on the flag.

        Args:
            flag (str): Flag to parse the records

        Returns:
            list: List of models based on the flag
        """
        logger.info(f"Parsing records with flag: {click.style(flag, fg='magenta')}")

        if flag not in self.flags:
            raise InvalidFlag(flag, self.flags)

        # Heart rate records have additional metadata (motionContext)
        if flag == "HKQuantityTypeIdentifierHeartRate":
            return [
                HeartRateData(
                    **{
                        **rec.attrib,
                        **{"motionContext": rec.find("MetadataEntry").attrib["value"]},
                    }
                )
                for rec in self.records[flag]
            ]
        else:
            return [HealthData(**rec.attrib) for rec in self.records[flag]]

    def _get_dates(self, models: list) -> set[date]:
        """
        Get unique month and year combinations from the models.

        Args:
            models (list): List of models

        Returns:
            set[datetime.date]: Set of dates (year, month, day)
        """
        return {rec.start_date.date for rec in models}

    def _map_record_keys_to_flags(self) -> dict[str, set]:
        """
        Map record keys (e.g. `unit`, `value`, `creationDate`) for each flag.

        For example:

        ```python
        HKCategoryTypeIdentifierAppleStandHour         {('type', 'sourceName', 'sourceVersion', 'device', ...)}
        HKCategoryTypeIdentifierAudioExposureEvent     {('type', 'sourceName', 'sourceVersion', 'device', ...)}
        HKDataTypeSleepDurationGoal                    {('type', 'sourceName', 'sourceVersion', 'unit', ...)}
        HKQuantityTypeIdentifierActiveEnergyBurned     {('type', 'sourceName', 'sourceVersion', 'device', ...)}
        ```

        Returns:
            dict[str, set]: Dictionary with flags as keys and set of record keys as values
        """
        return {
            flag: {tuple(rec.attrib.keys()) for rec in self.records[flag]}
            for flag in self.flags
        }

    @timeit
    def get_flag_records(
        self, flag: str | list[str]
    ) -> ParsedData | dict[str, ParsedData]:
        """
        Get parsed data based on the given flag.

        Args:
            flag (str | list[str]): Flag to parse the records (e.g., `"HKQuantityTypeIdentifierHeartRate"`)

        Returns:
            ParsedData | dict[str, ParsedData]: Parsed data based on the flag(s)
        """

        def _get_parsed_data(flag: str) -> ParsedData:
            sources = self.get_sources(flag=flag)
            models = self._build_models(flag=flag)
            dates = self._get_dates(models=models)
            records = pd.DataFrame([model.model_dump() for model in models])
            return ParsedData(flag=flag, sources=sources, dates=dates, records=records)

        if isinstance(flag, str):
            return _get_parsed_data(flag=flag)

        elif isinstance(flag, list):
            return {f: _get_parsed_data(flag=f) for f in flag}

    def get_sources(self, flag: str | None = None) -> list[str] | dict[str, list[str]]:
        """
        Get sources for each flag or for a given flag.

        Args:
            flag (str, optional): Get sources for the given flag, defaults to None

        Returns:
            list[str] | dict[str, list[str]]: Dictionary with flags as keys and list of sources as values
        """
        if flag:
            return sorted({rec.attrib["sourceName"] for rec in self.records[flag]})
        else:
            return {
                flag: sorted({rec.attrib["sourceName"] for rec in self.records[flag]})
                for flag in self.flags
            }

    @staticmethod
    def write_csv(data: ParsedData, filename: str) -> None:
        """
        Write the parsed data to a CSV file.

        Args:
            data (ParsedData): Parsed data to write to CSV file
            filename (str): Filename to write the data

        Raises:
            RecordsMissing: No records to write to CSV file
            FileTypeInvalid: File type is incorrect
        """
        if data.records is None:
            raise MissingRecords

        filepath = Path(filename)
        if filepath.suffix != ".csv":
            raise InvalidFileFormat(filepath.suffix)

        data.records.to_csv(filepath, index=False)
        logger.info(f"Parsed data written to {filepath}")

    @timeit
    def export(self, dir_name: str) -> None:
        """
        Export all parsed data to multiple CSV files.

        Args:
            dir_name (str): Directory name to export the CSV files
        """
        export_dir = Path(dir_name)
        export_dir.mkdir(exist_ok=True)

        logger.info(f"Exporting parsed data to {export_dir}...")

        for n, flag in enumerate(self.flags):
            try:
                parsed = self.get_flag_records(flag=flag)
            except:
                logger.error(f"Error parsing {flag=}")
                continue

            filename = f"{dir_name}/{flag}.csv"
            self.write_csv(data=parsed, filename=filename)

            logger.info(f"Exported {n+1}/{len(self.flags)} flags to {filename}")



from datetime import date
from itertools import compress
from pathlib import Path

from plotly import graph_objects as go
from plotly import subplots
from plotly.graph_objs import Figure

from apple_health_parser.config.definitions import AllowedImageFormats
from apple_health_parser.consts import FLAG_METADATA
from apple_health_parser.exceptions import InvalidImageFormat, MissingFlag
from apple_health_parser.models.parsed import ParsedData
from apple_health_parser.utils.preprocessor import Preprocessor


class ActivityOverview:
    def __init__(
        self,
        data: dict[str, ParsedData],
        year: int = date.today().year,
        source: str | None = None,
        title: bool = False,
    ) -> None:
        """
        Initialize the ActivityOverview object.

        Args:
            data (dict[str, ParsedData]): Dict with flag as key and ParsedData as value
            year (int, optional): Year, defaults to date.today().year
            source (str | None, optional):  Source, defaults to None
            title (bool, optional): Flag to include the plot title, defaults to False
        """
        self.src = source
        self.data = data
        self.year = year
        self.title = title
        self._validate()

        self.dataframes = {
            k: Preprocessor(data=v, year=year, source=source).get_dataframe()
            for k, v in data.items()
        }

    def _validate(self) -> None:
        """
        Validate the data.

        Raises:
            MissingFlag: Missing flag metadata
        """
        flags = [
            "HKQuantityTypeIdentifierActiveEnergyBurned",
            "HKQuantityTypeIdentifierAppleExerciseTime",
            "HKCategoryTypeIdentifierAppleStandHour",
        ]
        flags_not_in_data_bool: list[bool] = [flag not in self.data for flag in flags]
        if not any(flags_not_in_data_bool):
            flags_not_in_data: str = str(list(compress(flags, flags_not_in_data_bool)))
            raise MissingFlag(flags_not_in_data)

    def _get_figure(self) -> Figure:
        """
        Plot the activity overview.

        The activity overview includes the following flags:
            - Active Energy Burned (kcal)
            - Exercise Time (min)
            - Stand Time (min)

        Returns:
            Figure: Figure object
        """
        fig = subplots.make_subplots(rows=3, cols=1, shared_xaxes=True)

        for row, (flag, data) in enumerate(self.dataframes.items()):
            daily_sum = data.groupby("date")["value"].sum().reset_index()

            fig.append_trace(
                go.Bar(
                    x=daily_sum.date,
                    y=daily_sum.value,
                    marker_color=FLAG_METADATA[flag].color,
                    name=FLAG_METADATA[flag].name,
                ),
                row=row + 1,
                col=1,
            )

            fig.update_yaxes(
                title_text="<br>".join(
                    f"{FLAG_METADATA[flag].name} ({FLAG_METADATA[flag].unit})".split(
                        " "
                    )
                ),
                row=row + 1,
                col=1,
                title_font=dict(size=10),
            )

        fig.update_layout(
            title=f"Activity Overview {self.year}" if self.title else None,
            template="simple_white",
            showlegend=False,
        )

        return fig

    def plot(
        self, show: bool = True, save: bool = False, format: str = "png"
    ) -> Figure:
        """
        Plot the figure.

        Args:
            show (bool, optional): Whether to show the figure, defaults to True
            save (bool, optional): Whether to save the figure, defaults to False
            format (str, optional): Format of the file to be saved, defaults to "png"

        Raises:
            InvalidImageFormat: Image format is not allowed

        Returns:
            Figure: Figure object
        """
        figure = self._get_figure()

        if show:
            figure.show()

        if save:
            if format.upper() not in AllowedImageFormats.__members__:
                raise InvalidImageFormat(format)

            output_dir = Path("plots")
            output_dir.mkdir(exist_ok=True)
            filename = f"activity_overview_{self.year}"

            if format == "html":
                figure.write_html(output_dir / f"{filename}.html")
            else:
                figure.write_image(
                    file=output_dir / f"{filename}.{format}", format=format, scale=1
                )

        return figure



from apple_health_parser.config.definitions import (
    AllowedImageFormats,
    Operations,
    PlotType,
    get_flag_metadata,
)

FLAG_METADATA = get_flag_metadata()
ALLOWED_IMAGE_FORMATS = [fmt.value for fmt in AllowedImageFormats]
OPERATIONS = [op.value for op in Operations]
PLOT_TYPES = [ptype.value for ptype in PlotType]



import click
import pandas as pd

from apple_health_parser.interfaces.preprocessor_interface import PreprocessorInterface
from apple_health_parser.utils.logging import logger


class Preprocessor(PreprocessorInterface):
    """
    Preprocess the parsed data.

    Additional steps:
    - Validate `flag`, `source`, `operation`, and `year`
    - Filter the records for the given year (e.g. `2024`)
    - Apply the operation to the data if provided (e.g. `"mean"` or `"sum"`)
    """

    def get_dataframe(self) -> pd.DataFrame:
        """
        Get the preprocessed data in a DataFrame.

        Filter the records for the given year (e.g. `2024`).
        Apply the operation to the data if provided (e.g. `"mean"` or `"sum"`).

        For example, using the `"sum"` operation:

        ```bash
        date        value
        2024-01-01  10.0
        2024-01-02  15.0
        2024-01-03  7.0
        2024-02-01  14.0
        2024-02-02  15.0
        ...         ...
        ```

        Returns:
            pd.DataFrame: DataFrame with the preprocessed data
        """
        self.records = self.data.records

        # Filter by source (e.g. "Apple Watch" or "iPhone")
        if self.src:
            self.records = self.records[self.records.source_name == self.src]

        # Filter by year (e.g. 2024)
        self.records["date"] = self.records.start_date.dt.date
        self.records = self.records[
            self.records.date.apply(lambda x: x.year == self.year)
        ]

        # TODO: Handle other flags with special cases
        # Special case for HKQuantityTypeIdentifierOxygenSaturation (convert to percentage)
        if self.flag == "HKQuantityTypeIdentifierOxygenSaturation":
            self.records.value *= 100

        # Apply operation (e.g. "mean" or "sum")
        if self.oper:
            self.records = (
                self.records.groupby("date")["value"]
                .apply(getattr(pd.Series, self.oper))
                .round()
                .reset_index()
            )

            logger.info(
                f"Found {len(self.records)} records "
                f"(flag: {click.style(self.flag, fg='magenta')}, "
                f"operation: {click.style(self.oper, fg='blue')}, "
                f"year: {click.style(self.year, fg='green')})"
            )

            # Return heatmap data if requested
            if self.hmap:
                return self.get_heatmap(self.records)

        else:
            logger.info(
                f"Found {len(self.records)} records "
                f"(flag: {click.style(self.flag, fg='magenta')}, "
                f"year: {click.style(self.year, fg='green')})"
            )

        return self.records

    @staticmethod
    def get_heatmap(data: pd.DataFrame) -> pd.DataFrame:
        """
        Get the heatmap data.

        For example:

        ```bash
        day       1          2          3          4          5         ...
        month
        1        84.944151  76.080932  67.496200  88.942571  84.944151  ...
        2        84.944151  76.080932  67.496200  88.942571  84.944151  ...
        3        84.944151  76.080932  67.496200  88.942571  84.944151  ...
        ```

        Returns:
            pd.DataFrame: Heatmap data
        """
        data["month"] = data.date.apply(lambda x: x.month)
        data["day"] = data.date.apply(lambda x: x.day)
        heatmap = data.pivot(index="month", columns="day", values="value")
        return heatmap
